import os
import click
import tempfile
import yaml
import base64
import tarfile
import time
import shutil
import random

from pkg_resources import resource_string, resource_listdir
from subprocess import call
from .os_kube import OpenshiftKubeDeployer
from .more_objects import PersistentVolume
from .util import deepupdate
from pykube.objects import Pod
from pykube.config import KubeConfig

EDITOR = os.environ.get('EDITOR','vim')

@click.group()
@click.option("--config", default="~/.kube/config", help="kube config path", envvar="KUBE_CONFIG", type=click.Path())
@click.option("--context", default=None, help="kube context", envvar="KUBE_CONTEXT")
@click.option("--openshift-version", default="1.1.5", help="force openshift version (default 1.1.5)")
@click.option("--secure/--no-secure", default=False, help="enables https checking", envvar="KUBE_SESSION_SECURE")
@click.option("-y", help="auto-answer for all confirmations", is_flag=True)
@click.pass_context
def cli(ctx, config, context, secure, openshift_version, y):
    ctx.auto_confirm = y
    if ctx.auto_confirm:
        print("[warn] -y is set, will automatically confirm and proceed with actions.")

    ctx.kube_deployer = OpenshiftKubeDeployer(os.path.expanduser(config), context, secure)
    ctx.obj = ctx.kube_deployer
    ctx.obj.auto_confirm = y
    ctx.obj.os_version = openshift_version
    if ctx.obj.os_version.find("v") != 0:
        ctx.obj.os_version = "v" + ctx.obj.os_version
    ctx.obj.scripts_resource = __name__.split('.')[0] + ".resources.scripts"

@cli.command()
@click.pass_obj
def info(ctx):
    """Show cluster information"""
    exit(0 if ctx.init_with_checks() and ctx.consider_openshift_deployed else 1)

@cli.command()
@click.option("--persistent-volume", default="openshift-etcd1", help="Name of existing PersistentVolume of at least 2Gi size for storage")
@click.option("--create-volume/--no-create-volume", default=False, help="tell Kubernetes to create the volume (alpha feature)", envvar="OPENSHIFT_AUTOCREATE_VOLUME")
@click.option("--public-hostname", default=None, help="public hostname that will be DNSd to the public IP", envvar="OPENSHIFT_PUBLIC_DNS")
@click.option("--load-balancer/--no-load-balancer", default=True, help="use load balancer, otherwise node port", envvar="OPENSHIFT_CREATE_LOADBALANCER")
@click.option("--master-config-override", default=None, help="file with master-config.yaml overrides", envvar="OPENSHIFT_MASTER_CONFIG", type=click.Path(exists=True))
@click.option("--server-key", help="server.key from /srv/kubernetes/server.key", envvar="OPENSHIFT_SERVER_KEY_PATH", type=click.Path(exists=True))
@click.pass_obj
def deploy(ctx, persistent_volume, load_balancer, public_hostname, create_volume, master_config_override, server_key):
    """Deploy OpenShift to the cluster."""
    if not load_balancer and public_hostname != None:
        print("You must specify --load-balancer with --public-hostname, I can't map a public hostname without a load balancer.")
        exit(1)

    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if ctx.consider_openshift_deployed:
        print("I think OpenShift is already deployed. Use undeploy first to remove it before installing it again.")
        print("Consider if you really need a full redeploy. You can update without re-deploying!")
        exit(1)

    print()
    if "openshift" in ctx.namespace_names or "openshift-origin" in ctx.namespace_names:
        print("The namespaces 'openshift' and/or 'openshift-origin' exist, this indicates a potentially existing/broken install.")
        if ctx.auto_confirm:
            print("Auto confirm (-y) option set, clearing existing installation.")
        else:
            print("Really consider the decision you're about to make.")
            if not click.confirm("Do you want to clear the existing installation?"):
                print("Okay, cancelling.")
                exit(1)

        # Handle oddities with finalizers?
        # todo: delete "openshift" and "openshift-infra" namespaces
        ctx.delete_namespace_byname("openshift-origin")
        time.sleep(1)

    ctx.temp_dir = tempfile.mkdtemp()
    print("Preparing to execute deploy...")
    print("Deploy temp dir: " + ctx.temp_dir)

    # Setup the deploy state namespace
    ctx.cleanup_osdeploy_namespace()
    ctx.create_osdeploy_namespace()

    # Check the persistentvolume exists
    if not create_volume and ctx.find_persistentvolume(persistent_volume) == None:
        print(" [!] persistentvolume with name " + persistent_volume + " does not exist. Did you create it?")
        exit(1)

    # Create the namespaces
    ctx.create_namespace("openshift-origin")

    # Create the service
    if load_balancer:
        print("Will use load balancer type service.")
    else:
        print("Will use node port type service.")
    os_service = ctx.create_os_service(load_balancer)

    # Wait for it to be ready if it's a load balancer
    if load_balancer:
        print("Waiting for service load balancer IP to be allocated...")
        ctx.wait_for_loadbalancer(os_service)
    else:
        os_service.reload()

    internal_os_ip = os_service.obj["spec"]["clusterIP"]
    ctx.os_internal_ip = internal_os_ip

    if load_balancer:
        tmp = os_service.obj["status"]["loadBalancer"]["ingress"][0]
        external_os_ip = None
        external_is_hostname = False
        if "hostname" in tmp:
            external_os_ip = tmp["hostname"]
            external_is_hostname = True
        else:
            external_os_ip = tmp["ip"]
        ctx.os_external_ip = external_os_ip
        print("External OpenShift IP: " + external_os_ip)
    else:
        external_os_ip = internal_os_ip
        print("External OpenShift IP: nodes (node port)")

    print("Internal OpenShift IP: " + internal_os_ip)

    if public_hostname != None:
        print("You need to DNS map like this:")
        if external_is_hostname:
            print(public_hostname + ".\t300\tIN\tCNAME\t" + external_os_ip)
        else:
            print(public_hostname + ".\t300\tIN\tA\t" + external_os_ip)
        ctx.os_external_ip = public_hostname

    # Create a 'secret' containing the script to run to config.
    create_config_script = (resource_string(ctx.scripts_resource, 'create-config.sh'))

    # Build the secret
    create_config_secret_kv = {"create-config.sh": create_config_script}
    create_config_secret = ctx.build_secret("create-config-script", "openshift-deploy", create_config_secret_kv)
    create_config_secret.create()

    # Build the kubeconfig secret
    kubeconfig_secret_kv = {"kubeconfig": yaml.dump(ctx.config.doc).encode('ascii')}
    kubeconfig_secret = ctx.build_secret("kubeconfig", "openshift-deploy", kubeconfig_secret_kv)
    kubeconfig_secret.create()

    # Generate the openshift config by running a temporary pod on the cluster
    print("Generating openshift config via cluster...")
    conf_pod = ctx.build_config_pod(ctx.os_version)
    conf_pod.create()
    with open(ctx.temp_dir + "/config_bundle.tar.gz", 'wb') as f:
        conf_bundle = ctx.observe_config_pod(conf_pod)
        conf_bundle_data = base64.b64decode(conf_bundle)
        f.write(conf_bundle_data)
    conf_pod.delete()

    # Extract
    tar = tarfile.open(ctx.temp_dir + "/config_bundle.tar.gz")
    tar.extractall(ctx.temp_dir + "/config/")
    tar.close()

    # Move kubeconfig in
    with open(ctx.temp_dir + "/config/external-master.kubeconfig", 'w') as f:
        f.write(yaml.dump(ctx.config.doc))

    # Delete tarfile
    os.remove(ctx.temp_dir + "/config_bundle.tar.gz")

    # Do some processing on the master-config yaml
    conf = None
    with open(ctx.temp_dir + '/config/master-config.yaml') as f:
        conf = yaml.load(f)
    conf = ctx.fix_master_config(conf)

    # Write the serviceaccounts file again
    with open(server_key, 'r') as fs:
        with open(ctx.temp_dir + "/config/serviceaccounts.public.key", 'w') as fd:
            fd.write(fs.read())

    # Load patches if needed
    master_config_override_kv = None
    if master_config_override != None:
        print("Loading " + master_config_override + "...")
        with open(master_config_override, 'r') as f:
            master_config_override_kv = yaml.load(f)
        conf = deepupdate(conf, master_config_override_kv)

    # Write the fixed master config
    with open(ctx.temp_dir + "/config/master-config.yaml", 'w') as f:
        f.write(yaml.dump(conf, default_flow_style=False))

    # Allow the user to edit the openshift config last second
    print("Generated updated master-config.yaml.")
    if ctx.auto_confirm:
        print("Auto confirm (-y) option set, skipping master-config.yaml edit opportunity.")
    else:
        if click.confirm("Do you want to edit master-config.yaml?"):
            call([EDITOR, ctx.temp_dir + "/config/master-config.yaml"])

    # Cleanup a bit
    kubeconfig_secret.delete()
    create_config_secret.delete()

    # Serialize the config to a secret
    openshift_config_kv = {}
    for filen in os.listdir(ctx.temp_dir + "/config"):
        with open(ctx.temp_dir + "/config/" + filen, 'rb') as f:
            openshift_config_kv[filen] = f.read()
    openshift_config_secret = ctx.build_secret("openshift-config", "openshift-origin", openshift_config_kv)

    # Save the secret
    openshift_config_secret.create()

    # Starting etcd setup... build PersistentVolumeClaim
    etcd_pvc = ctx.build_pvc("openshift-etcd1", "openshift-origin", "2Gi", create_volume)
    etcd_pvc.create()

    # Create the etcd controller
    etcd_rc = ctx.build_etcd_rc("openshift-etcd1")
    etcd_svc = ctx.build_etcd_service()

    print("Creating etcd service...")
    etcd_svc.create()

    print("Creating etcd controller...")
    etcd_rc.create()

    print("Waiting for etcd pod to be created...")
    etcd_pod = None
    # Wait for the pod to exist
    while etcd_pod == None:
        etcd_pods = Pod.objects(ctx.api).filter(selector={"app": "etcd"}, namespace="openshift-origin").response["items"]
        if len(etcd_pods) < 1:
            time.sleep(0.5)
            continue
        etcd_pod = Pod(ctx.api, etcd_pods[0])

    # Wait for it to run
    ctx.wait_for_pod_running(etcd_pod)

    # Create the controller config
    print("Creating openshift replication controller...")
    openshift_rc = ctx.build_openshift_rc(ctx.os_version)
    openshift_rc.create()

    print("Waiting for openshift pod to be created...")
    openshift_pod = None
    # Wait for the pod to exist
    while openshift_pod == None:
        pods = Pod.objects(ctx.api).filter(namespace="openshift-origin", selector={"app": "openshift"}).response["items"]
        if len(pods) < 1:
            time.sleep(0.5)
            continue
        openshift_pod = Pod(ctx.api, pods[0])

    # Wait for it to run
    ctx.wait_for_pod_running(openshift_pod)

    print()
    print(" == OpenShift Deployed ==")
    print("External IP: " + ctx.os_external_ip)

    ctx.fetch_namespaces()
    ctx.cleanup_osdeploy_namespace()
    shutil.rmtree(ctx.temp_dir)

@cli.command()
@click.pass_obj
def undeploy(ctx):
    """Removes OpenShift from the cluster."""
    print("Preparing to remove OpenShift...")
    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if ctx.auto_confirm:
        print("Auto confirm (-y) option set, clearing existing installation.")
    else:
        print("Really consider the decision you're about to make.")
        if not click.confirm("Do you want to clear the existing installation?"):
            print("Okay, cancelling.")
            exit(1)
    ctx.delete_namespace_byname("openshift-origin")
    print("Note: namespaces with openshift finalizer will need to be manually deleted if desired.")
    print("See: https://github.com/paralin/openshift-under-kubernetes/blob/master/REMOVING_OPENSHIFT.md")

@cli.command()
@click.option("--config-output-dir", default=".", help="directory to write the openshift config", envvar="KUBE_CONFIG_OUTPUT_DIR", type=click.Path(exists=True), prompt=True)
@click.pass_obj
def getconfig(ctx, config_output_dir):
    """Writes the entire openshift config to a directory for inspection."""
    config_dir = config_output_dir
    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if not ctx.consider_openshift_deployed:
        print("I think OpenShift is not yet deployed. Use deploy first to create it.")
        exit(1)
    ctx.fetch_config_to_dir(config_dir)

@cli.command()
@click.option("--persistent-volume", default="openshift-registry", help="Name of existing PersistentVolume of at least 2Gi size for storage")
@click.option("--create-volume/--no-create-volume", default=False, help="tell Kubernetes to create the volume (alpha feature)", envvar="OPENSHIFT_AUTOCREATE_REGISTRY_VOLUME")
@click.option("--volume-size", default="10Gi", help="how big should the storage for the registry be", envvar="OPENSHIFT_REGISTRY_VOLUME_SIZE")
@click.option("--registry-image", default="openshift/origin-docker-registry", help="registry image name", envvar="OPENSHIFT_REGISTRY_IMAGE")
@click.pass_obj
def deployregistry(ctx, persistent_volume, create_volume, volume_size, registry_image):
    """Deploy an OpenShift Registry to the cluster."""
    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if not ctx.consider_openshift_deployed:
        print("OpenShift doesn't seem to be deployed.")
        print("Please deploy it first.")
        exit(1)

    # Build temporary object to check existence
    reg_rc = ctx.build_registry_rc("", "", "", "", "default", "registry-storage", "")
    if reg_rc.exists():
        print("The replication controller 'docker-registry' exists in 'default' namespace, this indicates a potentially existing/broken registry.")
        print("Refusing to continue, you should address this manually.")
        exit(1)

    print()

    ctx.temp_dir = tempfile.mkdtemp()
    print("Preparing to execute deploy...")
    print("Deploy temp dir: " + ctx.temp_dir)

    # Check the persistentvolume exists
    if not create_volume and ctx.find_persistentvolume(persistent_volume) == None:
        print(" [!] persistentvolume with name " + persistent_volume + " does not exist. Did you create it?")
        exit(1)

    # Fetch the config to the local dir
    ctx.fetch_config_to_dir(ctx.temp_dir)

    master_conf = None
    with open(ctx.temp_dir + "/master-config.yaml", 'r') as f:
        master_conf = yaml.load(f)
    internal_url = master_conf["oauthConfig"]["masterURL"]

    cluster_ca = None
    with open(ctx.temp_dir + "/ca.crt", 'r') as f:
        cluster_ca = f.read()

    client_key = None
    client_cert = None
    with open(ctx.temp_dir + "/openshift-registry.kubeconfig", 'r') as f:
        reg_conf = yaml.load(f)
        client_key = base64.b64decode(reg_conf["users"][0]["user"]["client-key-data"]).decode('ascii')
        client_cert = base64.b64decode(reg_conf["users"][0]["user"]["client-certificate-data"]).decode('ascii')

    # Create the namespaces
    #ctx.create_namespace("default")

    # Create the pvc
    reg_pvc = ctx.build_pvc("registry-storage", "default", volume_size, create_volume)
    if not reg_pvc.exists():
        reg_pvc.create()

    # Build the registry replication controller
    # I do this this way because I would prefer to not use a deployment here.
    reg_rc = ctx.build_registry_rc(cluster_ca, client_cert, client_key, internal_url, "default", "registry-storage", registry_image)
    # Needs a securitycontext that allows this
    if create_volume:
        reg_rc.obj["spec"]["template"]["spec"]["securityContext"] = {"fsGroup": 1001}
    # reg_rc.obj["spec"]["template"]["spec"]["securityContext"] = {"runAsUser": 0}
    reg_rc.create()

    # Create the service
    reg_svc = ctx.build_registry_svc("default")
    if not reg_svc.exists():
        reg_svc.create()

    print("Done!")
    shutil.rmtree(ctx.temp_dir)

@cli.command()
@click.pass_obj
def editconfig(ctx):
    """Interactively edits master-config.yaml"""
    ctx.temp_dir = tempfile.mkdtemp()
    if ctx.auto_confirm:
        print("Note: -y option is not supported for purely interactive commands.")
        ctx.auto_confirm = False

    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if not ctx.consider_openshift_deployed:
        print("I think OpenShift is not yet deployed. Use deploy first to create it.")
        exit(1)

    old_secret = ctx.fetch_config_to_dir(ctx.temp_dir)
    mc_path = ctx.temp_dir + "/master-config.yaml"
    if not os.path.exists(mc_path):
        print("Fetched config files but they don't contain master-config.yaml, something's wrong. Try getconfig.")
        shutil.rmtree(ctx.temp_dir)
        exit(1)

    last_mtime = os.path.getmtime(mc_path)
    print("Config files are at: " + ctx.temp_dir)
    print("Feel free to edit as you will...")
    print("Launching editor...")
    call([EDITOR, mc_path])
    now_mtime = os.path.getmtime(mc_path)
    if now_mtime == last_mtime:
        print("No changes made, exiting.")
        shutil.rmtree(ctx.temp_dir)
        exit(0)

    if not click.confirm("Do you want to upload the changed config files?"):
        print("Okay, cancelling.")
        shutil.rmtree(ctx.temp_dir)
        exit(0)

    print("Preparing to upload config files...")
    # Serialize the config to a secret
    openshift_config_kv = {}
    for filen in os.listdir(ctx.temp_dir):
        with open(ctx.temp_dir + "/" + filen, 'rb') as f:
            openshift_config_kv[filen] = f.read()
    openshift_config_secret = ctx.build_secret("openshift-config", "openshift-origin", openshift_config_kv)
    openshift_config_secret._original_obj = old_secret.obj

    print("Attempting to patch secret...")
    openshift_config_secret.update()
    print("Updates applied.")

    if not click.confirm("Do you want to restart openshift to apply the changes?"):
        print("Okay, I'm done. Have a nice day!")
        shutil.rmtree(ctx.temp_dir)
        exit(0)

    print("Restarting openshift pod...")
    try:
        pods = Pod.objects(ctx.api).filter(namespace="openshift-origin", selector={"app": "openshift"}).response["items"]
        if len(pods) >= 1:
            openshift_pod = Pod(ctx.api, pods[0])
            print("Deleting pod " + openshift_pod.obj["metadata"]["name"] + "...")
            openshift_pod.delete()
    except:
        print("Something went wrong restarting openshift, do it yourself please!")

    shutil.rmtree(ctx.temp_dir)

@cli.command()
@click.option("--username", default=None, prompt=True, help="username to alter", envvar="OPENSHIFT_TARGET_USERNAME")
@click.option("--role", prompt=True, default="cluster-admins", help="role to add", envvar="OPENSHIFT_TARGET_ROLE")
@click.pass_obj
def addclusterrole(ctx, username, role):
    """Adds a CLUSTER role to a user by directly querying the cluster."""
    if role != "cluster-admins":
        print("[warn] It's really not advisable to assign most cluster-wide roles to users.")

    admin_username = username
    ctx.temp_dir = tempfile.mkdtemp()
    if not ctx.escalate_admin_kubeconfig(ctx.temp_dir):
        print("Unable to escalate permissions using admin.kubeconfig, make sure it's valid.")
        exit(1)

    # We have admin permissions in openshift now.
    shutil.rmtree(ctx.temp_dir)

    # Attempt to query the users list
    # PyKube doesn't really support OpenShift so let's build urls ourself
    user_list = ctx.get_openshift_users()
    known_uid = []
    user = None
    for usr in user_list:
        name = usr["metadata"]["name"]
        known_uid.append(name)
        if name == admin_username:
            user = usr
    if user == None:
        print("Unable to find user '" + admin_username + "' in username list.")
        print("Known users: " + ", ".join(known_uid))
        exit(1)

    # User exists, check his roles
    if user["groups"] == None:
        user["groups"] = ["none"]
    print("Found user " + admin_username + ", groups: " + ", ".join(user["groups"]))

    # Add role
    cluster_role_bindings = ctx.get_openshift_cluster_rolebindings()
    mod_role = None
    for rolex in cluster_role_bindings:
        if rolex["metadata"]["name"] == role:
            mod_role = rolex
            break
    if mod_role == None:
        print("Unable to find cluster role named '" + role + "', try adding an s to the end of it.")
        exit(1)

    # Check the usernames
    if mod_role["userNames"] == None:
        mod_role["userNames"] = []
    unames = mod_role["userNames"]
    if admin_username in unames:
        print("User " + admin_username + " already in role binding " + mod_role["metadata"]["name"])
        print("Skipping update...")
    else:
        unames.append(admin_username)
        # patch the role binding
        print("Patching role binding...")
        ctx.put_openshift_cluster_rolebinding(mod_role)

    print("Done.")

@cli.command()
@click.option("--command", default="openshift ex diagnostics", help="command to run")
@click.pass_obj
def execute(ctx, command):
    """Executes any of the OpenShift commands inside a pod with the admin kubeconfig mounted."""
    if not ctx.init_with_checks():
        print("Failed cursory checks, exiting.")
        exit(1)

    if not ctx.consider_openshift_deployed:
        print("I think OpenShift is not yet deployed. Use deploy first to create it.")
        exit(1)

    # First grab the config to a folder
    ctx.temp_dir = tempfile.mkdtemp()
    ctx.fetch_config_to_dir(ctx.temp_dir)

    # Read the admin.kubeconfig
    admincfg = None
    with open(ctx.temp_dir + "/admin.kubeconfig", 'r') as f:
        admincfg = f.read()
    shutil.rmtree(ctx.temp_dir)

    # Create the pod
    exec_pod = ctx.build_execute_pod(command, admincfg)
    exec_pod.create()

    print()
    print("Remember to delete this pod after it's done executing.")
    print("You can clean up all execute pods with this command:")
    print("kubectl delete -l purpose=exec-command --namespace=openshift-origin")
    print()
    print("To view the result, use the command:")
    print("kubectl logs --namespace=openshift-origin -f " + exec_pod.obj["metadata"]["name"])

def main():
    cli()



def deepupdate(original, update):
    """
    Recursively update a dict.
    """
    for key, value in original.iteritems():
        if key not in update:
            update[key] = value
        elif isinstance(value, dict):
            deepupdate(value, update[key])
    return update



import os
import pykube
import requests
import time
import json
import base64
import copy
import yaml
import random
import string

from pykube.config import KubeConfig
from pykube.http import HTTPClient
from pykube.objects import Pod, Namespace, Service, ReplicationController, Secret
from .more_objects import PersistentVolume, PersistentVolumeClaim

def random_string(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

'''
Deploys OpenShift to Kubernetes
'''
class OpenshiftKubeDeployer:
    def __init__(self, config_path, context_override, enable_secure):
        self.config_path = config_path
        self.context_override = context_override
        self.enable_secure = enable_secure
        pass

    def init_with_checks(kube_deployer):
        print("Loading kube config...")
        if not kube_deployer.load_and_check_config():
            print("Unable to load/validate config.")
            return False

        print("Checking connectivity...")
        if not kube_deployer.fetch_namespaces():
            print("Connectivity looks bad, fix your connection!")
            return False

        print("Everything looks good, proceeding.")
        print()
        print("Collecting some initial cluster info...")
        kube_deployer.fetch_info(skip_namespaces=True)
        kube_deployer.print_openshift_basic_status()
        return True

    def load_and_check_config(self):
        if not os.path.exists(self.config_path):
            print("Config does not exist at path " + self.config_path + "!")
            return False
        try:
            self.config = KubeConfig.from_file(self.config_path)
        except:
            print("Config at path " + self.config_path + " failed to validate!")
            return False
        # Check current context
        if self.context_override != None:
            if self.context_override not in self.config.contexts:
                print("Context override " + self.context_override + " not in list of contexts.")
                return False
            self.config.set_current_context(self.context_override)
        elif self.config.current_context == None:
            print("Context not set, not sure which to use.")
            return False

        curr_ctx = self.config.contexts[self.config.current_context]
        self.api = HTTPClient(self.config)
        if not self.enable_secure:
            print('[note] we are in insecure mode, disabling warnings')
            requests.packages.urllib3.disable_warnings()
            self.api.session.verify = False
        return True

    def fetch_namespaces(self):
        self.namespace_list = Namespace.objects(self.api).all().response["items"]
        print("Currently there are " + str(len(self.namespace_list)) + " namespaces in the cluster.")
        if len(self.namespace_list) <= 0:
            return False
        self.namespace_names = []
        for ns in self.namespace_list:
            self.namespace_names.append(ns["metadata"]["name"])
        return True

    def fetch_openshift_setup(self):
        # Check if we have the openshift namespaces
        self.has_openshift_ns = "openshift-origin" in self.namespace_names
        # Check the replication controller
        try:
            self.openshift_rc = ReplicationController.objects(self.api).filter(namespace="openshift-origin", selector={"app": "openshift"}).response["items"][0]
        except Exception as ex:
            print(ex)
            self.openshift_rc = None
        self.consider_openshift_deployed = self.has_openshift_ns and self.openshift_rc != None

    def print_openshift_basic_status(kube_deployer):
        print("It looks like we " + ("do" if kube_deployer.has_openshift_ns else "do not") + " have the openshift-origin namespace, and " + ("do" if kube_deployer.openshift_rc != None else "do not") + " have a working ReplicationController.")
        print("Currently I " + ("do" if kube_deployer.consider_openshift_deployed else "do not") + " consider this a complete OpenShift deployment.")

    def fetch_info(self, skip_namespaces=False):
        if not skip_namespaces:
            self.fetch_namespaces()
        self.fetch_openshift_setup()

    def cleanup_osdeploy_namespace(self):
        for ns in self.namespace_list:
            if ns["metadata"]["name"] == "openshift-deploy":
                print("Cleaning up old openshift-deploy namespace.")
                self.delete_namespace_byname("openshift-deploy", True)
                return

    def build_namespace(self, name):
        return Namespace(self.api,
        {
            "metadata":
            {
                "name": name
            },
            "spec": {}
        })

    def create_namespace(self, name):
        print("Creating " + name + " namespace...")
        self.build_namespace(name).create()

    def delete_namespace_byname(self, name, wait_for_delete=False):
        print("Deleting " + name + " namespace...")
        try:
            self.build_namespace(name).delete()
        except Exception as ex:
            print("Ignoring potential error with delete: " + ex)
        if wait_for_delete:
            print("Waiting for " + name + " to terminate...")
        while wait_for_delete:
            wait_for_delete = len(Namespace.objects(self.api).filter(selector={"name": "openshift-deploy"}).response["items"]) != 0

    def create_osdeploy_namespace(self):
        self.create_namespace("openshift-deploy")


    def build_secret(self, name, namespace, kv):
        print("Creating " + name + " secret...")
        kv = copy.copy(kv)
        for k in kv:
            kv[k] = base64.b64encode(kv[k]).decode('ascii')
        return Secret(self.api,
        {
            "metadata":
            {
                "name": name,
                "namespace": namespace
            },
            "data": kv
        })

    def build_config_pod(self, os_version=""):
        if len(os_version) == 0:
            os_version = self.os_version

        print("Creating config generation pod...")
        return Pod(self.api,
        {
            "metadata":
            {
                "name": "generate-config",
                "labels":
                {
                    "purpose": "generate-config"
                },
                "namespace": "openshift-deploy"
            },
            "spec":
            {
                "containers":
                [{
                    "name": "generate-config",
                    "image": "openshift/origin:" + os_version,
                    "imagePullPolicy": "Always",
                    "command": ["/bin/bash"],
                    "args": ["/etc/config_secret_script/create-config.sh"],
                    "ports": [],
                    "env":
                    [{
                        "name": "OPENSHIFT_INTERNAL_ADDRESS",
                        "value": "https://" + self.os_internal_ip
                    }, {
                        "name": "OPENSHIFT_EXTERNAL_ADDRESS",
                        "value": "https://" + self.os_external_ip
                    }, {
                        "name": "ETCD_ADDRESS",
                        "value": "http://etcd:4001"
                    }],
                    "volumeMounts":
                    [{
                        "mountPath": "/etc/config_secret_script",
                        "name": "config-secret-script",
                        "readOnly": True
                    }, {
                        "mountPath": "/etc/kube_config",
                        "name": "kube-config",
                        "readOnly": True
                    }]
                }],
                "volumes":
                [{
                    "name": "config-secret-script",
                    "secret": {"secretName": "create-config-script"}
                }, {
                    "name": "kube-config",
                    "secret": {"secretName": "kubeconfig"}
                }],
                "restartPolicy": "Never"
            }
        })

    def build_execute_pod(self, command, admin_conf, os_version=""):
        if len(os_version) == 0:
            os_version = self.os_version

        print("Creating command execution pod...")
        name = "oskube-execute-" + random_string(4).lower()
        command = "mkdir -p ~/.kube/ && echo \"$ADMIN_KUBECONFIG\" > ~/.kube/config && " + command
        return Pod(self.api,
        {
            "metadata":
            {
                "name": name,
                "labels":
                {
                    "purpose": "exec-command"
                },
                "namespace": "openshift-origin"
            },
            "spec":
            {
                "containers":
                [{
                    "name": "exec-command",
                    "image": "openshift/origin:" + os_version,
                    "imagePullPolicy": "Always",
                    "command": ["/bin/bash"],
                    "args": ["-c", command],
                    "ports": [],
                    "env":
                    [{
                        "name": "ADMIN_KUBECONFIG",
                        "value": admin_conf
                    }]
                }],
                "restartPolicy": "Never"
            }
        })

    def wait_for_pod_succeed(self, pod):
        print("Waiting for " + pod.obj["metadata"]["name"] + " pod to finish...")
        while True:
            pod.reload()
            obj = pod.obj
            stat = obj["status"]["phase"]
            if stat in ["Running", "Pending"]:
                time.sleep(0.5)
            elif stat == "Succeeded":
                break
            else:
                raise Exception("Unexpected pod phase " + stat)

    def wait_for_pod_running(self, pod):
        print("Waiting for " + pod.obj["metadata"]["name"] + " pod to start...")
        while True:
            pod.reload()
            obj = pod.obj
            stat = obj["status"]["phase"]
            if stat in ["Pending"]:
                time.sleep(0.5)
            elif stat == "Running":
                break
            elif stat == "Succeeded":
                raise Exception("Unknown pod phase " + stat + ", expected this pod to be long-running.")
            else:
                raise Exception("Unknown pod phase " + stat)

    def observe_config_pod(self, pod, wait_for_start=True):
        if wait_for_start:
            self.wait_for_pod_succeed(pod)

        print("Checking logs...")
        url = self.api.url + "/api/v1/namespaces/openshift-deploy/pods/" + pod.obj["metadata"]["name"] + "/log"
        txt = self.api.session.get(url).text
        print("Retreived config bundle successfully.")
        return txt

    def build_os_service(self, use_load_balancer):
        return Service(self.api,
        {
            "metadata":
            {
                "name": "openshift",
                "namespace": "openshift-origin"
            },
            "spec":
            {
                "ports":
                [{
                    "name": "web",
                    "port": 443,
                    "targetPort": "web"
                }],
                "selector":
                {
                    "app": "openshift",
                    "tier": "backend"
                },
                "type": "LoadBalancer" if use_load_balancer else "NodePort"
            }
        })

    def create_os_service(self, use_load_balancer):
        print("Creating 'openshift' service...")
        srv = self.build_os_service(use_load_balancer)
        srv.create()
        return srv

    def wait_for_loadbalancer(self, service):
        while service.obj["spec"]["type"] == "LoadBalancer" \
            and ("loadBalancer" not in service.obj["status"] \
            or "ingress" not in service.obj["status"]["loadBalancer"] \
            or len(service.obj["status"]["loadBalancer"]["ingress"]) < 1 \
            or (("ip" not in service.obj["status"]["loadBalancer"]["ingress"][0] or len(service.obj["status"]["loadBalancer"]["ingress"][0]["ip"]) < 2)) and (("hostname" not in service.obj["status"]["loadBalancer"]["ingress"][0] or len(service.obj["status"]["loadBalancer"]["ingress"][0]["hostname"]) < 2))):
            time.sleep(0.5)
            service.reload()

    def fix_master_config(self, ya):
        # Fix ca
        # ya["kubeletClientInfo"]["ca"] = "ca.crt"
        ya["kubeletClientInfo"]["ca"] = ""
        ya["kubeletClientInfo"]["certFile"] = "master.kubelet-client.crt"
        ya["kubeletClientInfo"]["keyFile"] = "master.kubelet-client.key"
        ya["kubeletClientInfo"]["port"] = 10250

        # Fix etcd client
        ya["etcdClientInfo"]["ca"] = "etcd.server.crt"
        ya["etcdClientInfo"]["certFile"] = "master.etcd-client.crt"
        ya["etcdClientInfo"]["keyFile"] = "master.etcd-client.key"

        # Fix serviceAccountConfig
        ya["serviceAccountConfig"]["publicKeyFiles"] = ["serviceaccounts.public.key"]

        # Fix path to kubeconfig
        ya["masterClients"]["externalKubernetesKubeConfig"] = "external-master.kubeconfig"

        # Fix bind address
        ya["assetConfig"]["servingInfo"]["bindAddress"] = "0.0.0.0:443"
        ya["servingInfo"]["bindAddress"] = "0.0.0.0:443"

        return ya

    def build_pvc(self, name, namespace, size, create_volume):
        if not create_volume:
            return PersistentVolumeClaim(self.api,
            {
                "metadata":
                {
                    "name": name,
                    "namespace": namespace
                },
                "spec":
                {
                    "accessModes": ["ReadWriteOnce"],
                    "resources": {"requests": {"storage": size}}
                }
            })
        # See https://github.com/kubernetes/kubernetes/issues/22792#issuecomment-195563296
        else:
            return PersistentVolumeClaim(self.api,
            {
                "metadata":
                {
                    "name": name,
                    "annotations":
                    {
                        "volume.alpha.kubernetes.io/storage-class": "foo"
                    },
                    "namespace": namespace
                },
                "spec":
                {
                    "accessModes": ["ReadWriteOnce"],
                    "resources": {"requests": {"storage": size}}
                }
            })

    def build_etcd_rc(self, pvcn):
        return ReplicationController(self.api,
        {
            "metadata":
            {
                "name": "etcd",
                "namespace": "openshift-origin"
            },
            "spec":
            {
                "replicas": 1,
                "selector": {"app": "etcd", "purpose": "etcd"},
                "template":
                {
                    "metadata":
                    {
                        "labels": {"app": "etcd", "purpose": "etcd"}
                    },
                    "spec":
                    {
                        "containers":
                        [{
                            "name": "etcd",
                            "image": "openshift/etcd-20-centos7",
                            "ports": [{"containerPort": 4001, "name": "client"}, {"containerPort": 7001, "name": "server"}],
                            "command": ["/usr/local/bin/etcd"],
                            "args": ["--data-dir", "/var/lib/etcd", "--name", "openshift-etcd1"],
                            "env":
                            [{
                                "name": "ETCD_NUM_MEMBERS",
                                "value": "1"
                            }, {
                                "name": "ETCD_LISTEN_CLIENT_URLS",
                                "value": "http://0.0.0.0:2379,http://0.0.0.0:4001"
                            }],
                            "volumeMounts":
                            [{
                                "mountPath": "/var/lib/etcd",
                                "name": "etcd-storage"
                            }]
                        }],
                        "volumes":
                        [{
                            "name": "etcd-storage",
                            "persistentVolumeClaim": {"claimName": pvcn}
                        }],
                        "restartPolicy": "Always",
                        "dnsPolicy": "ClusterFirst"
                    }
                }
            }
        })

    def build_etcd_service(self):
        return Service(self.api,
        {
            "metadata":
            {
                "name": "etcd",
                "namespace": "openshift-origin"
            },
            "spec":
            {
                "ports":
                [{
                    "name": "client",
                    "port": 4001,
                    "targetPort": "client"
                }, {
                    "name": "server",
                    "port": 7001,
                    "targetPort": "server"
                }],
                "selector":
                {
                    "app": "etcd",
                    "purpose": "etcd"
                },
                "type": "ClusterIP"
            }
        })

    def build_registry_svc(self, namespace):
        return Service(self.api,
        {
            "metadata":
            {
                "labels":
                {
                    "app": "registry",
                    "role": "registry",
                    "tier": "backend",
                    # Not sure if this label is important
                    "docker-registry": "default"
                },
                "name": "docker-registry",
                "namespace": namespace
            },
            "spec":
            {
                "ports":
                [{
                    "name": "registry",
                    "targetPort": "registry",
                    "port": 5000,
                    "protocol": "TCP"
                }],
                "selector":
                {
                    "docker-registry": "default"
                }
            }
        })

    def build_registry_rc(self, ca_data, client_cert_data, client_key_data, server, namespace, pvcn, registry_image):
        return ReplicationController(self.api,
        {
            "metadata":
            {
                "labels":
                {
                    "app": "registry",
                    "role": "registry",
                    "tier": "backend",
                    # Not sure if this label is important
                    "docker-registry": "default"
                },
                "name": "docker-registry",
                "namespace": namespace
            },
            "spec":
            {
                "replicas": 1,
                "selector":
                {
                    "app": "registry",
                    "role": "registry",
                    "tier": "backend",
                    "docker-registry": "default"
                },
                "template":
                {
                    "metadata":
                    {
                        # I believe selector is filled with this on default
                        # Consider removing the extra info
                        "labels":
                        {
                            "app": "registry",
                            "role": "registry",
                            "tier": "backend",
                            "docker-registry": "default"
                        }
                    },
                    "spec":
                    {
                        "containers":
                        [{
                            "name": "registry",
                            "image": registry_image + ":" + self.os_version,
                            "imagePullPolicy": "IfNotPresent",
                            "livenessProbe":
                            {
                                "failureThreshold": 3,
                                "httpGet":
                                {
                                    "path": "/healthz",
                                    "port": 5000,
                                    "scheme": "HTTP"
                                },
                                "initialDelaySeconds": 10,
                                "periodSeconds": 10,
                                "successThreshold": 1,
                                "timeoutSeconds": 5
                            },
                            "ports": [{"containerPort": 5000, "protocol": "TCP", "name": "registry"}],
                            "volumeMounts": [{"mountPath": "/registry", "name": "registry-storage"}],
                            "env":
                            [{
                                "name": "OPENSHIFT_CA_DATA",
                                "value": ca_data
                            }, {
                                "name": "OPENSHIFT_CERT_DATA",
                                "value": client_cert_data
                            }, {
                                "name": "OPENSHIFT_INSECURE",
                                "value": "true" if server.startswith("http://") else "false"
                            }, {
                                "name": "OPENSHIFT_KEY_DATA",
                                "value": client_key_data
                            }, {
                                "name": "OPENSHIFT_MASTER",
                                "value": server
                            }, {
                                "name": "KUBERNETES_MASTER",
                                "value": server
                            }, {
                                # I'm guessing this tricks the registry into binding to port 5000
                                # Without actually setting an address
                                "name": "REGISTRY_HTTP_ADDR",
                                "value": ":5000"
                            }, {
                                "name": "REGISTRY_HTTP_NET",
                                "value": "tcp"
                            }, {
                                "name": "REGISTRY_HTTP_SECRET",
                                "value": random_string(10)
                            }]
                        }],
                        "volumes":
                        [{
                            "name": "registry-storage",
                            "persistentVolumeClaim": {"claimName": pvcn}
                        }],
                        "restartPolicy": "Always",
                        "dnsPolicy": "ClusterFirst"
                    }
                }
            }
        })

    def build_openshift_rc(self, os_version):
        return ReplicationController(self.api,
        {
            "metadata":
            {
                "name": "openshift",
                "namespace": "openshift-origin",
                "labels": {"app": "openshift", "tier": "backend"}
            },
            "spec":
            {
                "replicas": 1,
                "selector": {"app": "openshift", "tier": "backend"},
                "template":
                {
                    "metadata":
                    {
                        "labels": {"app": "openshift", "tier": "backend"}
                    },
                    "spec":
                    {
                        "containers":
                        [{
                            "image": "openshift/origin:" + os_version,
                            "name": "origin",
                            "args": ["start", "master", "--config=/config/master-config.yaml"],
                            "ports": [{"containerPort": 443, "name": "web"}],
                            "volumeMounts": [{"mountPath": "/config", "name": "config", "readOnly": True}]
                        }],
                        "volumes":
                        [{
                            "name": "config",
                            "secret": {"secretName": "openshift-config"}
                        }]
                    }
                }
            }
        })

    def find_persistentvolume(self, pvname):
        print("Checking for persistentvolume " + pvname + "...")
        try:
            url = self.api.url + "/api/v1/persistentvolumes/" + pvname
            res = self.api.session.get(url)
            if res.status_code != 200:
                return None
            return PersistentVolume(self.api, json.loads(res.text))
        except:
            return None

    def fetch_config_to_dir(self, config_dir):
        print("Fetching config files to dir " + config_dir)
        config_secret = self.build_secret("openshift-config", "openshift-origin", {})
        config_secret.reload()
        config_secret_kv = config_secret.obj["data"]
        print("Got config with " + str(len(config_secret_kv)) + " files.")

        # deserialize base64
        for k in config_secret_kv:
            config_secret_kv[k] = base64.b64decode(config_secret_kv[k]).decode('ascii')

        # write files
        for k in config_secret_kv:
            print("Writing " + k + " to " + config_dir)
            with open(config_dir + "/" + k, 'w') as f:
                f.write(config_secret_kv[k])

        print("Done fetching config.")
        return config_secret

    def escalate_admin_kubeconfig(self, temp_dir):
        # Fist we need to fetch the config
        ctx = self
        if not self.init_with_checks():
            print("Failed cursory checks.")
            return False

        if not ctx.consider_openshift_deployed:
            print("I think OpenShift is not yet deployed. Use deploy first to create it.")
            return False

        secret = ctx.fetch_config_to_dir(temp_dir)
        conf_path = temp_dir + "/admin.kubeconfig"
        mconf_path = temp_dir + "/master-config.yaml"
        if not os.path.exists(conf_path) or not os.path.exists(mconf_path):
            print("Fetched config files but they don't contain admin.kubeconfig, something's wrong. Try getconfig.")
            exit(1)

        print("Escalating to system:admin via admin.kubeconfig...")
        ctx.config_path = conf_path

        # Do an initial load to verify that it is valid
        self.context_override = None
        if not self.load_and_check_config():
            print("Unable to load/validate config.")
            return False

        # Pick the context for the public facing endpoint
        master_conf = None
        with open(mconf_path, 'r') as f:
            master_conf = yaml.load(f)

        pubUrl = master_conf["assetConfig"]["masterPublicURL"]
        cctx = self.config.current_context
        found_ctx = None
        for ctx in self.config.contexts:
            clun = self.config.contexts[ctx]["cluster"]
            if not clun in self.config.clusters:
                print("[warn] Cluster " + clun + " not valid, but in context " + ctx)
                continue

            # Grab the cluster
            clu = self.config.clusters[clun]
            if clu == None:
                continue
            if clu["server"] == pubUrl:
                print("Found context " + ctx + " for public-facing endpoint " + pubUrl)
                found_ctx = ctx
                break

        if found_ctx == None:
            print("[warn] Unable to find a context matching public url " + pubUrl + ", continuing with " + cctx)
        else:
            cctx = found_ctx
        self.context_override = cctx

        print("Re-checking connection with escalated permissions...")
        # This time we will override the context
        if not self.init_with_checks():
            print("Failed checks, make sure admin.kubeconfig is valid.")
            return False

        # We are now using the OpenShift admin connection.
        return True

    # Checks the OpenShift users list
    def get_openshift_users(self):
        url = self.api.url + "/oapi/v1/users"
        res = self.api.session.get(url).text
        return json.loads(res)["items"]

    # Checks the openshift cluster role binding list
    def get_openshift_cluster_rolebindings(self):
        url = self.api.url + "/oapi/v1/clusterrolebindings"
        res = self.api.session.get(url).text
        return json.loads(res)["items"]

    def put_openshift_cluster_rolebinding(self, rb):
        url = self.api.url + "/oapi/v1/clusterrolebindings/" + rb["metadata"]["name"]
        self.api.session.put(url, json=rb)



from pykube.objects import NamespacedAPIObject

class PersistentVolume(NamespacedAPIObject):

    version = "v1"
    endpoint = "persistentvolumes"

class PersistentVolumeClaim(NamespacedAPIObject):

    version = "v1"
    endpoint = "persistentvolumeclaims"
